"""bluepyopt.parameters tests"""

"""
Copyright (c) 2016-2020, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

# pylint:disable=W0612

import nose.tools as nt
from nose.plugins.attrib import attr

import bluepyopt


@attr('unit')
def test_parameters_init():
    """bluepyopt.parameters: test Parameter init"""

    param = bluepyopt.parameters.Parameter(name='test')
    nt.assert_is_instance(param, bluepyopt.parameters.Parameter)
    nt.assert_equal(param.name, 'test')


@attr('unit')
def test_parameters_fields():
    """bluepyopt.parameters: test Parameter fields"""

    param = bluepyopt.parameters.Parameter(name='test')

    nt.assert_equal(param.lower_bound, None)
    nt.assert_equal(param.upper_bound, None)

    param.freeze(5)
    nt.assert_raises(Exception, setattr, param, "value", 5)

    param = bluepyopt.parameters.Parameter(name='test', bounds=[2, 5])
    nt.assert_raises(ValueError, param.freeze, 1)


@attr('unit')
def test_parameters_str():
    """bluepyopt.parameters: test Parameter str conversion"""

    param = bluepyopt.parameters.Parameter(name='test')

    nt.assert_equal(str(param), 'test: value = None')

    param.freeze(5.5)

    nt.assert_equal(str(param), 'test: value = 5.5')


@attr('unit')
def test_MetaListEqualParameter_init():
    """bluepyopt.parameters: test MetaListEqualParameter init"""

    sub_params = [
        bluepyopt.parameters.Parameter(
            name='sub1', value=1), bluepyopt.parameters.Parameter(
            name='sub2', value=2)]

    nt.assert_equal(sub_params[0].value, 1)
    nt.assert_equal(sub_params[1].value, 2)

    param = bluepyopt.parameters.MetaListEqualParameter(
        name='param', value=0, frozen=True, sub_parameters=sub_params)
    nt.assert_is_instance(param, bluepyopt.parameters.Parameter)
    nt.assert_is_instance(param, bluepyopt.parameters.MetaListEqualParameter)

    nt.assert_equal(param.name, 'param')
    nt.assert_equal(param.sub_parameters[0].name, 'sub1')
    nt.assert_equal(param.sub_parameters[1].name, 'sub2')

    nt.assert_equal(param.value, 0)
    nt.assert_equal(sub_params[0].value, 0)
    nt.assert_equal(sub_params[1].value, 0)


@attr('unit')
def test_MetaListEqualParameter_freeze_unfreeze():
    """bluepyopt.parameters: test MetaListEqualParameter freeze and unfreeze"""

    sub_params = [
        bluepyopt.parameters.Parameter(
            name='sub1', value=1), bluepyopt.parameters.Parameter(
            name='sub2', value=2)]

    param = bluepyopt.parameters.MetaListEqualParameter(
        name='param', sub_parameters=sub_params)

    nt.assert_equal(param.value, None)
    nt.assert_equal(sub_params[0].value, 1)
    nt.assert_equal(sub_params[1].value, 2)

    param.freeze(0)

    nt.assert_equal(param.value, 0)
    nt.assert_equal(sub_params[0].value, 0)
    nt.assert_equal(sub_params[1].value, 0)

    param.unfreeze()

    sub_params[0].freeze(1)
    nt.assert_raises(Exception, param.freeze, 0)


@attr('unit')
def test_MetaListEqualParamete_str():
    """bluepyopt.parameters: test MetaListEqualParamete str conversion"""

    sub_params = [
        bluepyopt.parameters.Parameter(
            name='sub1', value=1), bluepyopt.parameters.Parameter(
            name='sub2', value=2)]

    param = bluepyopt.parameters.MetaListEqualParameter(
        name='param', sub_parameters=sub_params)

    nt.assert_equal(
        str(param),
        'param (sub_params: sub1: value = None,sub2: value = None): '
        'value = None')

    param.freeze(5.5)

    nt.assert_equal(
        str(param),
        'param (sub_params: sub1: value = 5.5,sub2: value = 5.5): '
        'value = 5.5')



"""Protocol classes"""

"""
Copyright (c) 2016-2020, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

# pylint: disable=W0511

import collections

# TODO: maybe find a better name ? -> sweep ?
import logging
logger = logging.getLogger(__name__)

from . import locations
from . import simulators


class Protocol(object):

    """Class representing a protocol (stimulus and recording)."""

    def __init__(self, name=None):
        """Constructor

        Args:
            name (str): name of the feature
        """

        self.name = name


class SequenceProtocol(Protocol):

    """A protocol consisting of a sequence of other protocols"""

    def __init__(self, name=None, protocols=None):
        """Constructor

        Args:
            name (str): name of this object
            protocols (list of Protocols): subprotocols this protocol
                consists of
        """
        super(SequenceProtocol, self).__init__(name)
        self.protocols = protocols

    def run(
            self,
            cell_model,
            param_values,
            sim=None,
            isolate=None,
            timeout=None):
        """Instantiate protocol"""

        responses = collections.OrderedDict({})

        for protocol in self.protocols:

            # Try/except added for backward compatibility
            try:
                response = protocol.run(
                    cell_model=cell_model,
                    param_values=param_values,
                    sim=sim,
                    isolate=isolate,
                    timeout=timeout)
            except TypeError as e:
                if "unexpected keyword" in str(e):
                    response = protocol.run(
                        cell_model=cell_model,
                        param_values=param_values,
                        sim=sim,
                        isolate=isolate)
                else:
                    raise

            key_intersect = set(
                response.keys()).intersection(set(responses.keys()))
            if len(key_intersect) != 0:
                raise Exception(
                    'SequenceProtocol: one of the protocols (%s) is trying to '
                    'add already existing keys to the response: %s' %
                    (protocol.name, key_intersect))

            responses.update(response)

        return responses

    def subprotocols(self):
        """Return subprotocols"""

        subprotocols = collections.OrderedDict({self.name: self})

        for protocol in self.protocols:
            subprotocols.update(protocol.subprotocols())

        return subprotocols

    def __str__(self):
        """String representation"""

        content = 'Sequence protocol %s:\n' % self.name

        content += '%d subprotocols:\n' % len(self.protocols)
        for protocol in self.protocols:
            content += '%s\n' % str(protocol)

        return content


class SweepProtocol(Protocol):

    """Sweep protocol"""

    def __init__(
            self,
            name=None,
            stimuli=None,
            recordings=None,
            cvode_active=None):
        """Constructor

        Args:
            name (str): name of this object
            stimuli (list of Stimuli): Stimulus objects used in the protocol
            recordings (list of Recordings): Recording objects used in the
                protocol
            cvode_active (bool): whether to use variable time step
        """

        super(SweepProtocol, self).__init__(name)
        self.stimuli = stimuli
        self.recordings = recordings
        self.cvode_active = cvode_active

    @property
    def total_duration(self):
        """Total duration"""

        return max([stimulus.total_duration for stimulus in self.stimuli])

    def subprotocols(self):
        """Return subprotocols"""

        return collections.OrderedDict({self.name: self})

    def _run_func(self, cell_model, param_values, sim=None):
        """Run protocols"""

        try:
            cell_model.freeze(param_values)
            cell_model.instantiate(sim=sim)

            self.instantiate(sim=sim, icell=cell_model.icell)

            try:
                sim.run(self.total_duration, cvode_active=self.cvode_active)
            except (RuntimeError, simulators.NrnSimulatorException):
                logger.debug(
                    'SweepProtocol: Running of parameter set {%s} generated '
                    'an exception, returning None in responses',
                    str(param_values))
                responses = {recording.name:
                             None for recording in self.recordings}
            else:
                responses = {
                    recording.name: recording.response
                    for recording in self.recordings}

            self.destroy(sim=sim)

            cell_model.destroy(sim=sim)

            cell_model.unfreeze(param_values.keys())

            return responses
        except BaseException:
            import sys
            import traceback
            raise Exception(
                "".join(
                    traceback.format_exception(*sys.exc_info())))

    def run(
            self,
            cell_model,
            param_values,
            sim=None,
            isolate=None,
            timeout=None):
        """Instantiate protocol"""

        if isolate is None:
            isolate = True

        if isolate:
            def _reduce_method(meth):
                """Overwrite reduce"""
                return (getattr, (meth.__self__, meth.__func__.__name__))

            import copyreg
            import types
            copyreg.pickle(types.MethodType, _reduce_method)
            import pebble
            from concurrent.futures import TimeoutError

            if timeout is not None:
                if timeout < 0:
                    raise ValueError("timeout should be > 0")

            with pebble.ProcessPool(max_workers=1, max_tasks=1) as pool:
                tasks = pool.schedule(self._run_func, kwargs={
                    'cell_model': cell_model,
                    'param_values': param_values,
                    'sim': sim},
                    timeout=timeout)
                try:
                    responses = tasks.result()
                except TimeoutError:
                    logger.debug('SweepProtocol: task took longer than '
                                 'timeout, will return empty response '
                                 'for this recording')
                    responses = {recording.name:
                                 None for recording in self.recordings}
        else:
            responses = self._run_func(cell_model=cell_model,
                                       param_values=param_values,
                                       sim=sim)
        return responses

    def instantiate(self, sim=None, icell=None):
        """Instantiate"""

        for stimulus in self.stimuli:
            stimulus.instantiate(sim=sim, icell=icell)

        for recording in self.recordings:
            try:
                recording.instantiate(sim=sim, icell=icell)
            except locations.EPhysLocInstantiateException:
                logger.debug(
                    'SweepProtocol: Instantiating recording generated '
                    'location exception, will return empty response for '
                    'this recording')

    def destroy(self, sim=None):
        """Destroy protocol"""

        for stimulus in self.stimuli:
            stimulus.destroy(sim=sim)

        for recording in self.recordings:
            recording.destroy(sim=sim)

    def __str__(self):
        """String representation"""

        content = '%s:\n' % self.name

        content += '  stimuli:\n'
        for stimulus in self.stimuli:
            content += '    %s\n' % str(stimulus)

        content += '  recordings:\n'
        for recording in self.recordings:
            content += '    %s\n' % str(recording)

        return content


class StepProtocol(SweepProtocol):

    """Protocol consisting of step and holding current"""

    def __init__(
            self,
            name=None,
            step_stimulus=None,
            holding_stimulus=None,
            recordings=None,
            cvode_active=None):
        """Constructor

        Args:
            name (str): name of this object
            step_stimulus (list of Stimuli): Stimulus objects used in protocol
            recordings (list of Recordings): Recording objects used in the
                protocol
            cvode_active (bool): whether to use variable time step
        """

        super(StepProtocol, self).__init__(
            name,
            stimuli=[
                step_stimulus,
                holding_stimulus]
            if holding_stimulus is not None else [step_stimulus],
            recordings=recordings,
            cvode_active=cvode_active)

        self.step_stimulus = step_stimulus
        self.holding_stimulus = holding_stimulus

    @property
    def step_delay(self):
        """Time stimulus starts"""
        return self.step_stimulus.step_delay

    @property
    def step_duration(self):
        """Time stimulus starts"""
        return self.step_stimulus.step_duration



"""Morphology classes"""

"""
Copyright (c) 2016-2020, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

# pylint: disable=W0511

import os
import platform
import logging
from bluepyopt.ephys.base import BaseEPhys
from bluepyopt.ephys.serializer import DictMixin

logger = logging.getLogger(__name__)

# TODO define an addressing scheme


class Morphology(BaseEPhys):

    """Morphology class"""
    pass


class NrnFileMorphology(Morphology, DictMixin):

    """Morphology loaded from a file"""
    SERIALIZED_FIELDS = ('morphology_path', 'do_replace_axon', 'do_set_nseg',
                         'replace_axon_hoc', )

    def __init__(
            self,
            morphology_path,
            do_replace_axon=False,
            do_set_nseg=True,
            comment='',
            replace_axon_hoc=None,
            nseg_frequency=40,
            morph_modifiers=None,
            morph_modifiers_hoc=None):
        """Constructor

        Args:
            morphology_path (str): location of the file describing the
                morphology
            do_replace_axon (bool): Does the axon need to be replaced by an AIS
                stub with default function ?
            replace_axon_hoc (str): String replacement for the 'replace_axon'
                command in hoc  Must include 'proc replace_axon(){ ... }
                If None,the default replace_axon is used
            nseg_frequency (float): frequency of nseg
            do_set_nseg (bool): if True, it will use nseg_frequency
            morph_modifiers (list): list of functions to modify the icell
                with (sim, icell) as arguments
            morph_modifiers_hoc (list): list of hoc strings corresponding
                to morph_modifiers
        """
        name = os.path.basename(morphology_path)
        super(NrnFileMorphology, self).__init__(name=name, comment=comment)
        # TODO speed up loading of morphologies from files
        # Path to morphology
        self.morphology_path = morphology_path
        self.do_replace_axon = do_replace_axon
        self.do_set_nseg = do_set_nseg
        self.nseg_frequency = nseg_frequency
        self.morph_modifiers = morph_modifiers
        self.morph_modifiers_hoc = morph_modifiers_hoc

        if replace_axon_hoc is None:
            self.replace_axon_hoc = self.default_replace_axon_hoc
        else:
            self.replace_axon_hoc = replace_axon_hoc

    def __str__(self):
        """Return string representation"""

        return self.morphology_path

    def instantiate(self, sim=None, icell=None):
        """Load morphology"""

        logger.debug('Loading morphology %s', self.morphology_path)

        if not os.path.exists(self.morphology_path):
            raise IOError(
                'Morphology not found at \'%s\'' %
                self.morphology_path)

        sim.neuron.h.load_file('stdrun.hoc')
        sim.neuron.h.load_file('import3d.hoc')

        extension = self.morphology_path.split('.')[-1]

        if extension.lower() == 'swc':
            imorphology = sim.neuron.h.Import3d_SWC_read()
        elif extension.lower() == 'asc':
            imorphology = sim.neuron.h.Import3d_Neurolucida3()
        else:
            raise ValueError("Unknown filetype: %s" % extension)

        # TODO this is to get rid of stdout print of neuron
        # probably should be more intelligent here, and filter out the
        # lines we don't want

        imorphology.quiet = 1

        if platform.system() == 'Windows':
            sim.neuron.h.hoc_stdout('NUL')
        else:
            sim.neuron.h.hoc_stdout('/dev/null')

        imorphology.input(str(self.morphology_path))
        sim.neuron.h.hoc_stdout()

        morphology_importer = sim.neuron.h.Import3d_GUI(imorphology, 0)

        morphology_importer.instantiate(icell)

        # TODO Set nseg should be called after all the parameters have been
        # set
        # (in case e.g. Ra was changed)
        if self.do_set_nseg:
            self.set_nseg(icell)

        if self.do_replace_axon:
            self.replace_axon(sim=sim, icell=icell)

        if self.morph_modifiers is not None:
            for morph_modifier in self.morph_modifiers:
                morph_modifier(sim=sim, icell=icell)

    def destroy(self, sim=None):
        """Destroy morphology instantiation"""
        pass

    def set_nseg(self, icell):
        """Set the nseg of every section"""
        for section in icell.all:
            section.nseg = 1 + 2 * int(section.L / self.nseg_frequency)

    @staticmethod
    def replace_axon(sim=None, icell=None):
        """Replace axon"""

        nsec = len([sec for sec in icell.axonal])

        if nsec == 0:
            ais_diams = [1, 1]
        elif nsec == 1:
            ais_diams = [icell.axon[0].diam, icell.axon[0].diam]
        else:
            ais_diams = [icell.axon[0].diam, icell.axon[0].diam]
            # Define origin of distance function
            sim.neuron.h.distance(0, 0.5, sec=icell.soma[0])

            for section in icell.axonal:
                # If distance to soma is larger than 60, store diameter
                if sim.neuron.h.distance(1, 0.5, sec=section) > 60:
                    ais_diams[1] = section.diam
                    break

        for section in icell.axonal:
            sim.neuron.h.delete_section(sec=section)

        # Create new axon array
        sim.neuron.h.execute('create axon[2]', icell)

        for index, section in enumerate(icell.axon):
            section.nseg = 1
            section.L = 30
            section.diam = ais_diams[index]
            icell.axonal.append(sec=section)
            icell.all.append(sec=section)

        icell.axon[0].connect(icell.soma[0], 1.0, 0.0)
        icell.axon[1].connect(icell.axon[0], 1.0, 0.0)

        logger.debug('Replace axon with AIS')

    default_replace_axon_hoc = \
        '''
proc replace_axon(){ local nSec, D1, D2
  // preserve the number of original axonal sections
  nSec = sec_count(axonal)

  // Try to grab info from original axon
  if(nSec == 0) { //No axon section present
    D1 = D2 = 1
  } else if(nSec == 1) {
    axon[0] D1 = D2 = diam
  } else {
    axon[0] D1 = diam
    soma distance() //to calculate distance from soma
    forsec axonal{
      //if section is longer than 60um then store diam and exit from loop
      if(distance(0.5) > 60){
        D2 = diam
        break
      }
    }
  }

  // get rid of the old axon
  forsec axonal{
    delete_section()
  }

  create axon[2]

  axon[0] {
    L = 30
    diam = D1
    nseg = 1 + 2*int(L/40)
    all.append()
    axonal.append()
  }
  axon[1] {
    L = 30
    diam = D2
    nseg = 1 + 2*int(L/40)
    all.append()
    axonal.append()
  }
  nSecAxonal = 2
  soma[0] connect axon[0](0), 1
  axon[0] connect axon[1](0), 1
}
        '''



"""Test ephys model objects"""

import os
import tempfile
import contextlib

import nose.tools as nt
from nose.plugins.attrib import attr

from bluepyopt import ephys

sim = ephys.simulators.NrnSimulator()
TESTDATA_DIR = os.path.join(
    os.path.dirname(
        os.path.abspath(__file__)),
    'testdata')
simple_morphology_path = os.path.join(TESTDATA_DIR, 'simple.swc')
apic_morphology_path = os.path.join(TESTDATA_DIR, 'apic.swc')


@contextlib.contextmanager
def yield_blank_hoc(template_name):
    """Create blank hoc template"""
    hoc_template = ephys.models.CellModel.create_empty_template(template_name)
    temp_file = tempfile.NamedTemporaryFile(suffix='test_models')
    with temp_file as fd:
        fd.write(hoc_template.encode('utf-8'))
        fd.flush()
        yield temp_file.name


test_morph = ephys.morphologies.NrnFileMorphology(simple_morphology_path)


@attr('unit')
def test_create_empty_template():
    """ephys.models: Test creation of empty template"""
    template_name = 'FakeTemplate'
    hoc_template = ephys.models.CellModel.create_empty_template(template_name)
    sim.neuron.h(hoc_template)
    nt.assert_true(hasattr(sim.neuron.h, template_name))


@attr('unit')
def test_model():
    """ephys.models: Test Model class"""
    model = ephys.models.Model('test_model')
    model.instantiate(sim=None)
    model.destroy(sim=None)
    nt.assert_true(isinstance(model, ephys.models.Model))


@attr('unit')
def test_cellmodel():
    """ephys.models: Test CellModel class"""
    model = ephys.models.CellModel('test_model', morph=test_morph, mechs=[])

    nt.assert_equal(
        str(model),
        'test_model:\n  morphology:\n    %s\n  mechanisms:\n  params:\n' %
        simple_morphology_path)

    model.instantiate(sim=sim)
    model.destroy(sim=sim)
    nt.assert_true(isinstance(model, ephys.models.CellModel))


@attr('unit')
def test_cellmodel_namecheck():
    """ephys.models: Test CellModel class name checking"""

    # Test valid name
    for name in ['test3', 'test_3']:
        ephys.models.CellModel(name, morph=test_morph, mechs=[])

    # Test invalid names
    for name in ['3test', '', 'test$', 'test 3']:
        nt.assert_raises(
            TypeError,
            ephys.models.CellModel,
            name,
            morph=test_morph,
            mechs=[])


@attr('unit')
def test_load_hoc_template():
    """ephys.models: Test loading of hoc template"""

    template_name = 'test_load_hoc'
    hoc_string = ephys.models.CellModel.create_empty_template(template_name)
    ephys.models.HocCellModel.load_hoc_template(sim, hoc_string)
    nt.ok_(hasattr(sim.neuron.h, template_name))


@attr('unit')
def test_HocCellModel():
    """ephys.models: Test HOCCellModel class"""
    template_name = 'test_HocCellModel'
    hoc_string = ephys.models.CellModel.create_empty_template(template_name)
    hoc_cell = ephys.models.HocCellModel(
        'test_hoc_model', simple_morphology_path, hoc_string=hoc_string)
    hoc_cell.instantiate(sim)
    nt.ok_(hoc_cell.icell is not None)
    nt.ok_(hoc_cell.cell is not None)

    nt.ok_('simple.swc' in str(hoc_cell))

    # these should be callable, but don't do anything
    hoc_cell.freeze(None)
    hoc_cell.unfreeze(None)
    hoc_cell.check_nonfrozen_params(None)
    hoc_cell.params_by_names(None)

    hoc_cell.destroy(sim=sim)


@attr('unit')
def test_CellModel_create_empty_cell():
    """ephys.models: Test create_empty_cell"""
    template_name = 'create_empty_cell'
    cell = ephys.models.CellModel.create_empty_cell(template_name, sim)
    nt.assert_true(callable(cell))
    nt.assert_true(hasattr(sim.neuron.h, template_name))


@attr('unit')
def test_CellModel_create_hoc():
    """ephys.models: Test create_hoc"""

    morph0 = ephys.morphologies.NrnFileMorphology(
        simple_morphology_path,
        do_replace_axon=True)

    cell_model = ephys.models.CellModel('CellModel',
                                        morph=morph0,
                                        mechs=[],
                                        params=[])

    hoc_string = cell_model.create_hoc({})
    nt.assert_true('begintemplate CellModel' in hoc_string)
    nt.assert_true('proc replace_axon()' in hoc_string)
    cell_model_hoc = ephys.models.HocCellModel(
        'CellModelHOC',
        simple_morphology_path,
        hoc_string=hoc_string)

    nt.assert_true(isinstance(cell_model_hoc, ephys.models.HocCellModel))


@attr('unit')
def test_CellModel_destroy():
    """ephys.models: Test CellModel destroy"""
    morph0 = ephys.morphologies.NrnFileMorphology(simple_morphology_path)
    cell_model0 = ephys.models.CellModel('CellModel_destroy',
                                         morph=morph0,
                                         mechs=[],
                                         params=[])
    morph1 = ephys.morphologies.NrnFileMorphology(simple_morphology_path)
    cell_model1 = ephys.models.CellModel('CellModel_destroy',
                                         morph=morph1,
                                         mechs=[],
                                         params=[])

    nt.assert_true(not hasattr(sim.neuron.h, 'CellModel_destroy'))

    cell_model0.instantiate(sim=sim)
    nt.assert_true(hasattr(sim.neuron.h, 'CellModel_destroy'))
    nt.assert_equal(1, len(sim.neuron.h.CellModel_destroy))

    cell_model1.instantiate(sim=sim)
    nt.assert_equal(2, len(sim.neuron.h.CellModel_destroy))

    # make sure cleanup works
    cell_model0.destroy(sim=sim)
    nt.assert_equal(1, len(sim.neuron.h.CellModel_destroy))

    cell_model1.destroy(sim=sim)
    nt.assert_equal(0, len(sim.neuron.h.CellModel_destroy))


@attr('unit')
def test_metaparameter():
    """ephys.models: Test model with MetaParameter"""

    morph = ephys.morphologies.NrnFileMorphology(apic_morphology_path)

    dist = "({A} + {B} * math.exp({distance} * {C})) * {value}"

    scaler = ephys.parameterscalers.NrnSegmentSomaDistanceScaler(
        distribution=dist, dist_param_names=['A', 'B', 'C'])

    all_loc = ephys.locations.NrnSeclistLocation('all', 'all')

    paramA = ephys.parameters.MetaParameter('ParamA', scaler, 'A', -1)
    paramB = ephys.parameters.MetaParameter(
        'ParamB',
        scaler,
        'B',
        bounds=[1.0, 3.0])
    paramC = ephys.parameters.MetaParameter(
        'ParamC',
        obj=scaler,
        attr_name='C',
        value=0.003,
        frozen=True)

    cm = ephys.parameters.NrnRangeParameter(
        name='cm',
        param_name='cm',
        bounds=[.5, 1.5],
        value_scaler=scaler,
        locations=[all_loc])

    test_params = {'ParamA': -1, 'ParamB': 2.0, 'cm': 1.0}

    cell_model = ephys.models.CellModel('CellModel',
                                        morph=morph,
                                        mechs=[],
                                        params=[cm, paramA, paramB, paramC])

    nt.assert_raises(Exception,
                     cell_model.freeze,
                     {'ParamC': 2.0})

    cell_model.freeze(test_params)

    cell_model.instantiate(sim=sim)

    nt.assert_equal(scaler.eval_dist(1.0, 1.0),
                    '(-1 + 2.0 * math.exp(1 * 0.003)) * 1')

    nt.assert_almost_equal(
        scaler.scale(
            1.0,
            cell_model.icell.apic[0](.5),
            sim=sim),
        1.045510068328892)

    cell_model.unfreeze(param_names=['ParamA', 'ParamB'])

    hoc_code = cell_model.create_hoc(param_values=test_params)

    nt.assert_true('distribute_distance(CellRef.all, "cm", "(-1 + 2.0 * '
                   'exp(%.17g * 0.003)) * 1")' in hoc_code)

    cell_model.destroy(sim=sim)

    hoc_model = ephys.models.HocCellModel(
        'hoc_model', '.', hoc_string=hoc_code)
    hoc_model.instantiate(sim=sim)
    hoc_model.destroy()



"""Test ephys model objects"""

# pylint: disable=R0914

import os

import nose.tools as nt
from nose.plugins.attrib import attr

from bluepyopt import ephys

TESTDATA_DIR = os.path.join(
    os.path.dirname(
        os.path.abspath(__file__)),
    'testdata')

simple_morphology_path = os.path.join(TESTDATA_DIR, 'simple.swc')


@attr('unit')
def test_CellEvaluator_init():
    """ephys.evaluators: Test CellEvaluator init"""
    sim = ephys.simulators.NrnSimulator()

    model = ephys.models.CellModel('test_model', params=[])

    fitness_calc = ephys.objectivescalculators.ObjectivesCalculator()

    evaluator = ephys.evaluators.CellEvaluator(
        cell_model=model,
        param_names=[],
        fitness_calculator=fitness_calc,
        sim=sim)

    nt.assert_true(isinstance(evaluator, ephys.evaluators.CellEvaluator))
    nt.assert_equal(
        str(evaluator),
        'cell evaluator:\n  cell model:\n    test_model:\n  morphology:\n  '
        'mechanisms:\n  params:\n\n  fitness protocols:\n  '
        'fitness calculator:\n    objectives:\n\n')


@attr('unit')
def test_CellEvaluator_evaluate():
    """ephys.evaluators: Test CellEvaluator evaluate"""
    sim = ephys.simulators.NrnSimulator()

    simple_morph = ephys.morphologies.NrnFileMorphology(
        simple_morphology_path,
        do_replace_axon=True)

    all_loc = ephys.locations.NrnSeclistLocation('all', 'all')

    cm = ephys.parameters.NrnRangeParameter(
        name='cm',
        param_name='cm',
        bounds=[.5, 1.5],
        locations=[all_loc])

    cell_model = ephys.models.CellModel('CellModel',
                                        morph=simple_morph,
                                        mechs=[],
                                        params=[cm])

    soma_loc = ephys.locations.NrnSeclistCompLocation(
        name='soma_loc',
        seclist_name='somatic',
        sec_index=0,
        comp_x=.5)

    rec_soma = ephys.recordings.CompRecording(
        name='soma.v',
        location=soma_loc,
        variable='v')

    stim = ephys.stimuli.NrnSquarePulse(
        step_amplitude=0.1,
        step_delay=100.0,
        step_duration=100,
        total_duration=200,
        location=soma_loc)

    protocol = ephys.protocols.SweepProtocol(
        name='prot',
        stimuli=[stim],
        recordings=[rec_soma])

    mean = -65

    efeature = ephys.efeatures.eFELFeature(name='test_eFELFeature',
                                           efel_feature_name='voltage_base',
                                           recording_names={'': 'soma.v'},
                                           stim_start=100.0,
                                           stim_end=200.0,
                                           exp_mean=-65,
                                           exp_std=1)

    s_obj = ephys.objectives.SingletonObjective(
        'singleton',
        feature=efeature)

    fitness_calc = ephys.objectivescalculators.ObjectivesCalculator(
        objectives=[s_obj])

    evaluator = ephys.evaluators.CellEvaluator(
        cell_model=cell_model,
        param_names=['cm'],
        fitness_calculator=fitness_calc,
        fitness_protocols={'sweep': protocol},
        sim=sim)

    responses = protocol.run(cell_model, {'cm': 1.0}, sim=sim)
    feature_value = efeature.calculate_feature(responses)

    score = evaluator.evaluate([1.0])
    expected_score = abs(mean - feature_value)

    nt.assert_almost_equal(score, expected_score)

    score_dict = evaluator.objective_dict(score)

    nt.assert_almost_equal(score_dict['singleton'], expected_score)



"""Tests for ephys.efeatures"""

import os

import nose.tools as nt
from nose.plugins.attrib import attr

import bluepyopt.ephys as ephys


@attr('unit')
def test_EFeatureObjective():
    """ephys.objectives: Testing EFeatureObjective"""

    recording_names = {'': 'square_pulse_step1.soma.v'}

    mean = 1

    efeature = ephys.efeatures.eFELFeature(name='test_eFELFeature',
                                           efel_feature_name='voltage_base',
                                           recording_names=recording_names,
                                           stim_start=700,
                                           stim_end=2700,
                                           exp_mean=mean,
                                           exp_std=1)

    e_obj = ephys.objectives.EFeatureObjective(
        'singleton',
        features=[efeature])

    nt.assert_equal(e_obj.name, 'singleton')
    nt.assert_equal(e_obj.features, [efeature])

    response = ephys.responses.TimeVoltageResponse('mock_response')
    testdata_dir = os.path.join(
        os.path.dirname(
            os.path.abspath(__file__)),
        'testdata')
    response.read_csv(os.path.join(testdata_dir, 'TimeVoltageResponse.csv'))
    responses = {'square_pulse_step1.soma.v': response, }

    efeature_value = efeature.calculate_feature(responses)

    nt.assert_almost_equal(
        e_obj.calculate_feature_scores(responses),
        [abs(efeature_value - mean)])


@attr('unit')
def test_SingletonObjective():
    """ephys.objectives: Testing SingletonObjective"""

    recording_names = {'': 'square_pulse_step1.soma.v'}

    mean = 1

    efeature = ephys.efeatures.eFELFeature(name='test_eFELFeature',
                                           efel_feature_name='voltage_base',
                                           recording_names=recording_names,
                                           stim_start=700,
                                           stim_end=2700,
                                           exp_mean=mean,
                                           exp_std=1)

    s_obj = ephys.objectives.SingletonObjective(
        'singleton',
        feature=efeature)

    nt.assert_equal(s_obj.name, 'singleton')
    nt.assert_equal(s_obj.features, [efeature])
    nt.assert_equal(str(s_obj), '( %s )' % str(efeature))

    response = ephys.responses.TimeVoltageResponse('mock_response')
    testdata_dir = os.path.join(
        os.path.dirname(
            os.path.abspath(__file__)),
        'testdata')
    response.read_csv(os.path.join(testdata_dir, 'TimeVoltageResponse.csv'))
    responses = {'square_pulse_step1.soma.v': response, }

    efeature_value = efeature.calculate_feature(responses)

    nt.assert_almost_equal(
        s_obj.calculate_score(responses),
        abs(efeature_value - mean))


@attr('unit')
def test_MaxObjective():
    """ephys.objectives: Testing MaxObjective"""

    recording_names = {'': 'square_pulse_step1.soma.v'}

    mean = 1

    efeature1 = ephys.efeatures.eFELFeature(name='test_eFELFeature',
                                            efel_feature_name='voltage_base',
                                            recording_names=recording_names,
                                            stim_start=700,
                                            stim_end=2700,
                                            exp_mean=mean,
                                            exp_std=1)
    efeature2 = ephys.efeatures.eFELFeature(
        name='test_eFELFeature',
        efel_feature_name='steady_state_voltage',
        recording_names=recording_names,
        stim_start=700,
        stim_end=2700,
        exp_mean=mean,
        exp_std=1)

    max_obj = ephys.objectives.MaxObjective(
        'max',
        features=[efeature1, efeature2])

    nt.assert_equal(max_obj.name, 'max')
    nt.assert_equal(max_obj.features, [efeature1, efeature2])

    response = ephys.responses.TimeVoltageResponse('mock_response')
    testdata_dir = os.path.join(
        os.path.dirname(
            os.path.abspath(__file__)),
        'testdata')
    response.read_csv(os.path.join(testdata_dir, 'TimeVoltageResponse.csv'))
    responses = {'square_pulse_step1.soma.v': response, }

    efeature_value1 = efeature1.calculate_feature(responses)
    efeature_value2 = efeature2.calculate_feature(responses)

    nt.assert_almost_equal(
        max_obj.calculate_score(responses),
        max(abs(efeature_value1 - mean), abs(efeature_value2 - mean)))


@attr('unit')
def test_WeightedSumObjective():
    """ephys.objectives: Testing WeightedSumObjective"""

    recording_names = {'': 'square_pulse_step1.soma.v'}

    mean = 1
    std = 1

    efeature = ephys.efeatures.eFELFeature(name='test_eFELFeature',
                                           efel_feature_name='voltage_base',
                                           recording_names=recording_names,
                                           stim_start=700,
                                           stim_end=2700,
                                           exp_mean=mean,
                                           exp_std=std)

    weight = .5

    w_obj = ephys.objectives.WeightedSumObjective(
        'weighted',
        features=[efeature],
        weights=[weight])

    nt.assert_equal(w_obj.name, 'weighted')
    nt.assert_equal(w_obj.features, [efeature])
    nt.assert_equal(w_obj.weights, [weight])

    response = ephys.responses.TimeVoltageResponse('mock_response')
    testdata_dir = os.path.join(
        os.path.dirname(
            os.path.abspath(__file__)),
        'testdata')
    response.read_csv(os.path.join(testdata_dir, 'TimeVoltageResponse.csv'))
    responses = {'square_pulse_step1.soma.v': response, }

    efeature_value = efeature.calculate_feature(responses)

    nt.assert_almost_equal(
        w_obj.calculate_score(responses),
        abs(efeature_value - mean) * weight)

    nt.assert_raises(Exception, ephys.objectives.WeightedSumObjective,
                     'weighted',
                     features=[efeature],
                     weights=[1, 2])



"""bluepyopt.optimisations tests"""

import nose.tools as nt

import bluepyopt.optimisations
import bluepyopt.ephys.examples as examples

from nose.plugins.attrib import attr

import deap.tools


@attr('unit')
def test_DEAPOptimisation_constructor():
    "deapext.optimisation: Testing constructor DEAPOptimisation"

    optimisation = bluepyopt.deapext.optimisations.DEAPOptimisation(
        examples.simplecell.cell_evaluator, map_function=map)

    nt.assert_is_instance(
        optimisation,
        bluepyopt.deapext.optimisations.DEAPOptimisation)
    nt.assert_is_instance(
        optimisation.evaluator,
        bluepyopt.evaluators.Evaluator)

    nt.assert_raises(
        ValueError,
        bluepyopt.deapext.optimisations.DEAPOptimisation,
        examples.simplecell.cell_evaluator,
        selector_name='wrong')


@attr('unit')
def test_IBEADEAPOptimisation_constructor():
    "deapext.optimisation: Testing constructor IBEADEAPOptimisation"

    optimisation = bluepyopt.deapext.optimisations.IBEADEAPOptimisation(
        examples.simplecell.cell_evaluator, map_function=map)

    nt.assert_is_instance(
        optimisation,
        bluepyopt.deapext.optimisations.IBEADEAPOptimisation)


@attr('unit')
def test_DEAPOptimisation_run():
    "deapext.optimisation: Testing DEAPOptimisation run"

    optimisation = bluepyopt.optimisations.DEAPOptimisation(
        examples.simplecell.cell_evaluator, offspring_size=1)

    pop, hof, log, hist = optimisation.run(max_ngen=1)

    ind = [0.06007731830843009, 0.06508319290092013]
    nt.assert_equal(len(pop), 1)
    nt.assert_almost_equal(pop[0], ind)
    nt.assert_almost_equal(hof[0], ind)
    nt.assert_equal(log[0]['nevals'], 1)
    nt.assert_almost_equal(hist.genealogy_history[1], ind)


@attr('unit')
def test_DEAPOptimisation_run_from_parents():
    "deapext.optimisation: Testing DEAPOptimisation run using prior parents"

    optimisation = bluepyopt.optimisations.DEAPOptimisation(
        examples.simplecell.cell_evaluator, offspring_size=1)

    parent_population = [[0.060, 0.065]]
    pop, hof, log, hist = optimisation.run(max_ngen=0,
                                           parent_population=parent_population)

    nt.assert_equal(len(pop), 1)
    nt.assert_almost_equal(pop[0], parent_population[0])


@attr('unit')
def test_selectorname():
    "deapext.optimisation: Testing selector_name argument"

    # Test default value
    ibea_optimisation = bluepyopt.optimisations.DEAPOptimisation(
        examples.simplecell.cell_evaluator)
    nt.assert_equal(ibea_optimisation.selector_name, 'IBEA')

    # Test NSGA2 selector
    nsga2_optimisation = bluepyopt.optimisations.DEAPOptimisation(
        examples.simplecell.cell_evaluator, selector_name='NSGA2')

    nt.assert_equal(
        nsga2_optimisation.toolbox.select.func,
        deap.tools.emo.selNSGA2)

    # Test IBEA selector
    ibea_optimisation = bluepyopt.optimisations.DEAPOptimisation(
        examples.simplecell.cell_evaluator, selector_name='IBEA')

    nt.assert_equal(
        ibea_optimisation.toolbox.select.func,
        bluepyopt.deapext.tools.selIBEA)



"""
Mechanism classes

Theses classes represent mechanisms in the model
"""

"""
Copyright (c) 2016-2020, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

# pylint: disable=W0511

import logging

from . import base
from . import serializer

logger = logging.getLogger(__name__)

# TODO: use Location class to specify location


class Mechanism(base.BaseEPhys):

    """Base parameter class"""
    pass


class NrnMODMechanism(Mechanism, serializer.DictMixin):

    """Neuron mechanism"""

    SERIALIZED_FIELDS = (
        'name',
        'comment',
        'mod_path',
        'suffix',
        'locations',
        'preloaded',
    )

    def __init__(
            self,
            name,
            mod_path=None,
            suffix=None,
            locations=None,
            preloaded=True,
            deterministic=True,
            prefix=None,
            comment=''):
        """Constructor

        Args:
            name (str): name of this object
            mod_path (str): path to the MOD file (not used for the moment)
            suffix (str): suffix of this mechanism in the MOD file
            locations (list of Locations): a list of Location objects pointing
                to where this mechanism should be added to.
            preloaded (bool): should this mechanism be side-loaded by
                BluePyOpt, or was it already loaded and compiled by the user ?
                (not used for the moment)
            prefix (str): Deprecated. Use suffix instead.
        """

        super(NrnMODMechanism, self).__init__(name, comment)
        self.mod_path = mod_path
        self.suffix = suffix
        self.locations = locations
        self.preloaded = preloaded
        self.cell_model = None
        self.deterministic = deterministic

        if prefix is not None and suffix is not None:
            raise TypeError('NrnMODMechanism: it is not allowed to set both '
                            'prefix and suffix in constructor: %s %s' %
                            (self.prefix, self.suffix))
        elif prefix is not None:
            self.suffix = prefix

    def instantiate(self, sim=None, icell=None):
        """Instantiate"""

        for location in self.locations:
            isec_list = location.instantiate(sim=sim, icell=icell)
            for isec in isec_list:
                try:
                    isec.insert(self.suffix)
                except ValueError as e:
                    raise ValueError(str(e) + ': ' + self.suffix)
                self.instantiate_determinism(
                    self.deterministic,
                    icell,
                    isec,
                    sim)

        logger.debug(
            'Inserted %s in %s', self.suffix, [
                str(location) for location in self.locations])

    def instantiate_determinism(self, deterministic, icell, isec, sim):
        """Instantiate enable/disable determinism"""

        if 'Stoch' in self.suffix:
            setattr(
                isec,
                'deterministic_%s' %
                (self.suffix),
                1 if deterministic else 0)

            if not deterministic:
                # Set the seeds
                short_secname = sim.neuron.h.secname(sec=isec).split('.')[-1]
                for iseg in isec:
                    seg_name = '%s.%.19g' % (short_secname, iseg.x)
                    getattr(sim.neuron.h,
                            "setdata_%s" % self.suffix)(iseg.x, sec=isec)
                    seed_id1 = icell.gid
                    seed_id2 = self.hash_py(seg_name)
                    getattr(
                        sim.neuron.h,
                        "setRNG_%s" % self.suffix)(seed_id1, seed_id2)
        else:
            if not deterministic:
                # can't do this for non-Stoch channels
                raise TypeError(
                    'Deterministic can only be set to False for '
                    'Stoch channel, not %s' %
                    self.suffix)

    def destroy(self, sim=None):
        """Destroy mechanism instantiation"""

        pass

    def __str__(self):
        """String representation"""

        return "%s: %s at %s" % (
            self.name, self.suffix,
            [str(location) for location in self.locations])

    @staticmethod
    def hash_hoc(string, sim):
        """Calculate hash value of string in Python"""

        # Load hash function in hoc, only do this once
        if not hasattr(sim.neuron.h, 'hash_str'):
            sim.neuron.h(NrnMODMechanism.hash_hoc_string)

        return sim.neuron.h.hash_str(string)

    @staticmethod
    def hash_py(string):
        """Calculate hash value of string in Python"""

        hash_value = 0.0
        for char in string:
            # Multiplicative hash function using Mersenne prime close to 2^32
            hash_value = (hash_value * 31.0 + ord(char)) % (2.0 ** 31.0 - 1.0)

        return hash_value

    def generate_reinitrng_hoc_block(self):
        """"Create re_init_rng code blocks for this channel"""

        reinitrng_hoc_block = ''

        if 'Stoch' in self.suffix:
            # TODO this is dangerous, implicitely assumes type of location
            for location in self.locations:
                if self.deterministic:
                    reinitrng_hoc_block += \
                        '    forsec %(seclist_name)s { ' \
                        'deterministic_%(suffix)s = 1 }\n' % {
                            'seclist_name': location.seclist_name,
                            'suffix': self.suffix}
                else:
                    reinitrng_hoc_block += \
                        '    forsec %(seclist_name)s {%(mech_reinitrng)s' \
                        '    }\n' % {
                            'seclist_name': location.seclist_name,
                            'mech_reinitrng':
                            self.mech_reinitrng_block_template % {
                                'suffix': self.suffix}}

        return reinitrng_hoc_block

    @property
    def prefix(self):
        """Deprecated, prefix is now replaced by suffix"""

        return self.suffix

    @prefix.setter
    def prefix(self, value):
        """Deprecated, prefix is now replaced by suffix"""

        self.suffix = value

    hash_hoc_string = \
        """
func hash_str() {localobj sf strdef right
  sf = new StringFunctions()

  right = $s1

  n_of_c = sf.len(right)

  hash = 0
  char_int = 0
  for i = 0, n_of_c - 1 {
     sscanf(right, "%c", & char_int)
     hash = (hash * 31 + char_int) % (2 ^ 31 - 1)
     sf.right(right, 1)
  }

  return hash
}
"""

    reinitrng_hoc_string = """
proc re_init_rng() {localobj sf
    strdef full_str, name

    sf = new StringFunctions()

    if(numarg() == 1) {
        // We received a third seed
        channel_seed = $1
        channel_seed_set = 1
    } else {
        channel_seed_set = 0
    }

%(reinitrng_hoc_blocks)s
}
"""

    mech_reinitrng_block_template = """
        for (x, 0) {
            setdata_%(suffix)s(x)
            sf.tail(secname(), "\\\\.", name)
            sprint(full_str, "%%s.%%.19g", name, x)
            if (channel_seed_set) {
                setRNG_%(suffix)s(gid, hash_str(full_str), channel_seed)
            } else {
                setRNG_%(suffix)s(gid, hash_str(full_str))
            }
        }
"""


class NrnMODPointProcessMechanism(Mechanism):

    """Neuron mechanism"""

    def __init__(
            self,
            name,
            mod_path=None,
            suffix=None,
            locations=None,
            preloaded=True,
            comment=''):
        """Constructor

        Args:
            name (str): name of this object
            mod_path (str): path to the MOD file (not used for the moment)
            suffix (str): suffix of this mechanism in the MOD file
            locations (list of Locations): a list of Location objects pointing
                to compartments where this mechanism should be added to.
            preloaded (bool): should this mechanism be side-loaded by
                BluePyOpt, or was it already loaded and compiled by the user ?
                (not used for the moment)
        """

        super(NrnMODPointProcessMechanism, self).__init__(name, comment)
        self.mod_path = mod_path
        self.suffix = suffix
        self.locations = locations
        self.preloaded = preloaded
        self.cell_model = None
        self.pprocesses = None

    def instantiate(self, sim=None, icell=None):
        """Instantiate"""

        self.pprocesses = []
        for location in self.locations:
            icomp = location.instantiate(sim=sim, icell=icell)
            try:
                iclass = getattr(sim.neuron.h, self.suffix)
                self.pprocesses.append(iclass(icomp.x, sec=icomp.sec))
            except AttributeError as e:
                raise AttributeError(str(e) + ': ' + self.suffix)

        logger.debug(
            'Inserted %s at %s ', self.suffix, [
                str(location) for location in self.locations])

    def destroy(self, sim=None):
        """Destroy mechanism instantiation"""

        self.pprocesses = None

    def __str__(self):
        """String representation"""

        return "%s: %s at %s" % (
            self.name, self.suffix,
            [str(location) for location in self.locations])



"""Stimuli classes"""

"""
Copyright (c) 2016-2020, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

# pylint: disable=W0511

import logging
logger = logging.getLogger(__name__)


class Stimulus(object):

    """Stimulus protocol"""
    pass


class NrnCurrentPlayStimulus(Stimulus):

    """Current stimulus based on current amplitude and time series"""

    def __init__(self,
                 time_points=None,
                 current_points=None,
                 location=None):
        """Constructor

        Args:
            time_points(): time series (ms)
            current_points(): current series of injected current amplitudes(nA)
            location(Location): location of stimulus
        """

        super(NrnCurrentPlayStimulus, self).__init__()
        self.time_points = time_points
        self.current_points = current_points
        self.location = location
        self.total_duration = max(time_points)
        self.iclamp = None
        self.current_vec = None
        self.time_vec = None

    def instantiate(self, sim=None, icell=None):
        """Run stimulus"""

        icomp = self.location.instantiate(sim=sim, icell=icell)
        logger.debug(
            'Adding current play stimulus to %s', str(self.location))

        self.iclamp = sim.neuron.h.IClamp(
            icomp.x,
            sec=icomp.sec)
        self.current_vec = sim.neuron.h.Vector(self.current_points)
        self.time_vec = sim.neuron.h.Vector(self.time_points)
        self.iclamp.dur = self.total_duration
        self.iclamp.delay = 0
        self.current_vec.play(
            self.iclamp._ref_amp,  # pylint:disable=W0212
            self.time_vec,
            1,
            sec=icomp.sec)

    def destroy(self, sim=None):
        """Destroy stimulus"""

        self.iclamp = None
        self.time_vec = None
        self.current_vec = None

    def __str__(self):
        """String representation"""

        return "Current play at %s" % (self.location)


class NrnNetStimStimulus(Stimulus):

    """Current stimulus based on current amplitude and time series"""

    def __init__(self,
                 locations=None,
                 total_duration=None,
                 interval=None,
                 number=None,
                 start=None,
                 noise=0,
                 weight=1):
        """Constructor

        Args:
            location: synapse point process location to connect to
            interval: time between spikes (ms)
            number: average number of spikes
            start: most likely start time of first spike (ms)
            noise: fractional randomness (0 deterministic,
                   1 negexp interval distrubtion)
        """

        super(NrnNetStimStimulus, self).__init__()
        if total_duration is None:
            raise ValueError(
                'NrnNetStimStimulus: Need to specify a total duration')
        else:
            self.total_duration = total_duration

        self.locations = locations
        self.interval = interval
        self.number = number
        self.start = start
        self.noise = noise
        self.weight = weight
        self.connections = {}

    def instantiate(self, sim=None, icell=None):
        """Run stimulus"""

        for location in self.locations:
            self.connections[location.name] = []
            for synapse in location.instantiate(sim=sim, icell=icell):
                netstim = sim.neuron.h.NetStim()
                netstim.interval = self.interval
                netstim.number = self.number
                netstim.start = self.start
                netstim.noise = self.noise
                netcon = sim.neuron.h.NetCon(netstim, synapse)
                netcon.weight[0] = self.weight

                self.connections[location.name].append((netcon, netstim))

    def destroy(self, sim=None):
        """Destroy stimulus"""

        self.connections = None

    def __str__(self):
        """String representation"""

        return "Netstim at %s" % ','.join(
            location
            for location in self.locations) \
            if self.locations is not None else "Netstim"

# TODO Add 'current' to the name


class NrnSquarePulse(Stimulus):

    """Square pulse current clamp injection"""

    def __init__(self,
                 step_amplitude=None,
                 step_delay=None,
                 step_duration=None,
                 total_duration=None,
                 location=None):
        """Constructor

        Args:
            step_amplitude (float): amplitude (nA)
            step_delay (float): delay (ms)
            step_duration (float): duration (ms)
            total_duration (float): total duration (ms)
            location (Location): stimulus Location
        """

        super(NrnSquarePulse, self).__init__()
        self.step_amplitude = step_amplitude
        self.step_delay = step_delay
        self.step_duration = step_duration
        self.location = location
        self.total_duration = total_duration
        self.iclamp = None

    def instantiate(self, sim=None, icell=None):
        """Run stimulus"""

        icomp = self.location.instantiate(sim=sim, icell=icell)
        logger.debug(
            'Adding square step stimulus to %s with delay %f, '
            'duration %f, and amplitude %f',
            str(self.location),
            self.step_delay,
            self.step_duration,
            self.step_amplitude)

        self.iclamp = sim.neuron.h.IClamp(
            icomp.x,
            sec=icomp.sec)
        self.iclamp.dur = self.step_duration
        self.iclamp.amp = self.step_amplitude
        self.iclamp.delay = self.step_delay

    def destroy(self, sim=None):
        """Destroy stimulus"""

        self.iclamp = None

    def __str__(self):
        """String representation"""

        return "Square pulse amp %f delay %f duration %f totdur %f at %s" % (
            self.step_amplitude,
            self.step_delay,
            self.step_duration,
            self.total_duration,
            self.location)


class NrnRampPulse(Stimulus):

    """Ramp current clamp injection"""

    def __init__(self,
                 ramp_amplitude_start=None,
                 ramp_amplitude_end=None,
                 ramp_delay=None,
                 ramp_duration=None,
                 total_duration=None,
                 location=None):
        """Constructor

        Args:
            ramp_amplitude_start (float): amplitude at start of ramp (nA)
            ramp_amplitude_start (float): amplitude at end of ramp (nA)
            ramp_delay (float): delay of ramp (ms)
            ramp_duration (float): duration oframp (ms)
            total_duration (float): total duration (ms)
            location (Location): stimulus Location
        """

        super(NrnRampPulse, self).__init__()
        self.ramp_amplitude_start = ramp_amplitude_start
        self.ramp_amplitude_end = ramp_amplitude_end
        self.ramp_delay = ramp_delay
        self.ramp_duration = ramp_duration
        self.location = location
        self.total_duration = total_duration
        self.iclamp = None
        self.persistent = []  # TODO move this into higher abstract classes

    def instantiate(self, sim=None, icell=None):
        """Run stimulus"""

        icomp = self.location.instantiate(sim=sim, icell=icell)
        logger.debug(
            'Adding ramp stimulus to %s with delay %f, '
            'duration %f, amplitude at start %f and end %f',
            str(self.location),
            self.ramp_delay,
            self.ramp_duration,
            self.ramp_amplitude_start,
            self.ramp_amplitude_end
        )

        # create vector to store the times at which stim amp changes
        times = sim.neuron.h.Vector()
        # create vector to store to which stim amps over time
        amps = sim.neuron.h.Vector()

        # at time 0.0, current is 0.0
        times.append(0.0)
        amps.append(0.0)

        # until time ramp_delay, current is 0.0
        times.append(self.ramp_delay)
        amps.append(0.0)

        # at time ramp_delay, current is ramp_amplitude_start
        times.append(self.ramp_delay)
        amps.append(self.ramp_amplitude_start)

        # at time ramp_delay+ramp_duration, current is ramp_amplitude_end
        times.append(self.ramp_delay + self.ramp_duration)
        amps.append(self.ramp_amplitude_end)

        # after ramp, current is set 0.0
        times.append(self.ramp_delay + self.ramp_duration)
        amps.append(0.0)

        times.append(self.total_duration)
        amps.append(0.0)

        # create a current clamp
        self.iclamp = sim.neuron.h.IClamp(
            icomp.x,
            sec=icomp.sec)
        self.iclamp.dur = self.total_duration

        # play the above current amplitudes into the current clamp
        amps.play(self.iclamp._ref_amp, times, 1)  # pylint: disable=W0212

        # Make sure the following objects survive after instantiation
        self.persistent.append(times)
        self.persistent.append(amps)

    def destroy(self, sim=None):
        """Destroy stimulus"""

        # Destroy all persistent objects
        self.persistent = []
        self.iclamp = None

    def __str__(self):
        """String representation"""

        return "Ramp pulse amp_start %f amp_end %f delay %f duration %f " \
            "totdur %f at %s" % (
                self.ramp_amplitude_start,
                self.ramp_amplitude_end,
                self.ramp_delay,
                self.ramp_duration,
                self.total_duration,
                self.location)



"""Cell evaluator class"""

"""
Copyright (c) 2016-2020, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

from abc import abstractmethod


class Evaluator(object):

    """Evaluator class

    An Evaluator maps a set of parameter values to objective values
        Args:
            objectives (Objectives):
                The objectives that will be the output of the evaluator.
            params (Parameters):
                The parameters that will be evaluated.

        Attributes:
            objectives (Objectives):
                Objective objects.
            params (Objectives):
                Parameter objects.
    """

    def __init__(self, objectives=None, params=None):
        self.objectives = objectives
        self.params = params

    # TODO add evaluate_with_dicts
    @abstractmethod
    def evaluate_with_dicts(self, param_dict):
        """Evaluate parameter a parameter set (abstract).

        Args:
            params (dict with values Parameters, and keys parameter names):
                The parameter values to be evaluated.

        Returns:
            objectives (dict with values Parameters, and keys objective names):
                Dict of Objective with values calculated by the Evaluator.

        """

    @abstractmethod
    def evaluate_with_lists(self, params):
        """Evaluate parameter a parameter set (abstract).

        Args:
            params (list of Parameters):
                The parameter values to be evaluated.

        Returns:
            objectives (list of Objectives):
                List of Objectives with values calculated by the Evaluator.

        """



"""bluepyopt.stoppingCriteria tests"""

import nose.tools as nt

import bluepyopt.stoppingCriteria

from nose.plugins.attrib import attr

import deap.tools


@attr('unit')
def test_MaxNGen():
    """deapext.stoppingCriteria: Testing MaxNGen"""

    max_gen = 3
    criteria = bluepyopt.deapext.stoppingCriteria.MaxNGen(max_gen)

    nt.assert_equal(criteria.criteria_met, False)
    criteria.check({"gen": max_gen + 1})
    nt.assert_equal(criteria.criteria_met, True)
    criteria.reset()
    criteria.check({"gen": max_gen})
    nt.assert_equal(criteria.criteria_met, False)



"""Dummy cell model used for testing"""

import bluepyopt.ephys as ephys


class DummyCellModel1(ephys.models.Model):

    """Dummy cell model 1"""

    def __init__(self, name=None):
        """Constructor"""

        super(DummyCellModel1, self).__init__(name)
        self.persistent = []
        self.icell = None

    def freeze(self, param_values):
        """Freeze model"""
        pass

    def unfreeze(self, param_names):
        """Freeze model"""
        pass

    def instantiate(self, sim=None):
        """Instantiate cell in simulator"""

        class Cell(object):

            """Empty cell class"""

            def __init__(self):
                """Constructor"""
                self.soma = None
                self.somatic = None

        self.icell = Cell()

        self.icell.soma = [sim.neuron.h.Section(name='soma', cell=self.icell)]
        self.icell.apic = [
            sim.neuron.h.Section(
                name='apic1',
                cell=self.icell)]

        self.icell.somatic = sim.neuron.h.SectionList(
        )  # pylint: disable = W0201
        self.icell.somatic.append(sec=self.icell.soma[0])

        self.icell.apical = sim.neuron.h.SectionList()
        self.icell.apical.append(sec=self.icell.apic[0])

        self.persistent.append(self.icell)
        self.persistent.append(self.icell.soma[0])

        return self.icell

    def destroy(self, sim=None):
        """Destroy cell from simulator"""

        self.persistent = []



"""Get stats out of ipyparallel's tasks.db"""

from __future__ import print_function

"""
Copyright (c) 2016-2020, EPFL/Blue Brain Project

 This file is part of BluePyOpt <https://github.com/BlueBrain/BluePyOpt>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import sys
import os
import argparse
import sqlite3
import collections
import datetime
import dateutil.parser
import itertools

import numpy as np
import matplotlib.pyplot as plt


def get_engine_data(tasksdb_filename):
    """Main"""
    conn = sqlite3.connect(tasksdb_filename)

    cursor = conn.cursor()

    tasks = collections.defaultdict(list)

    SQL = 'SELECT started, completed, engine_uuid FROM "ipython-tasks";'

    for started, completed, engine_uuid in cursor.execute(SQL).fetchall():
        if started and completed:
            started = dateutil.parser.parse(started)
            completed = dateutil.parser.parse(completed)

            duration = (
                completed -
                started).total_seconds() if completed else None
            task = {'started': started,
                    'completed': completed,
                    'duration': duration,
                    'engine_uuid': engine_uuid}

            tasks[engine_uuid].append(task)

    if len(tasks) == 0:
        raise Exception("No completed tasks found in the db")

    engine_number_map = dict(zip(tasks.keys(), range(len(tasks.keys()))))
    return tasks, engine_number_map


def plot_usage(tasks, engine_number_map):
    """Plot usage stats"""

    plt.figure(figsize=(8, 8))
    for engine_uuid, task_list in tasks.items():
        engine_number = engine_number_map[engine_uuid]
        number_list = [engine_number] * len(task_list)
        start_list = [task['started'] for task in task_list]
        completed_list = [task['completed'] for task in task_list]
        plt.plot(
            [number_list, number_list],
            [start_list, completed_list], linewidth=10,
            solid_capstyle="butt")

    plt.xlim(min(engine_number_map.values()) - 1,
             max(engine_number_map.values()) + 1)
    plt.xlabel('Compute engine number')
    plt.ylabel('Compute time')

    idle_time, idle_perc = calculate_unused_compute(tasks)
    plt.title(
        'Cumulative idle time: %s, perc: %.2f %%' %
        (idle_time, idle_perc))


def plot_duration_histogram(tasks):
    """Plot duration histogram"""
    plt.figure(figsize=(8, 8))

    durations = np.fromiter((t['duration']
                             for task_list in tasks.values()
                             for t in task_list),
                            dtype=np.float)
    plt.hist(durations, 100)

    plt.xlabel('Duration (s)')
    plt.ylabel('Count')
    plt.title('Histogram of task execution')
    plt.grid(True)


def calculate_unused_compute(tasks):
    """Calculate unused compute time"""

    all_tasks = list(itertools.chain.from_iterable(tasks.values()))
    start_time = min(task['started'] for task in all_tasks)
    end_time = max(task['completed'] for task in all_tasks)
    total_compute_time = sum(
        (datetime.timedelta(seconds=task['duration'])
         for task in all_tasks), datetime.timedelta())

    n_of_engines = len(tasks.keys())

    idle_time = n_of_engines * (end_time - start_time) - total_compute_time
    idle_perc = 100 * \
        (idle_time.total_seconds() / total_compute_time.total_seconds())
    return idle_time, idle_perc


def run(arg_list):
    """Main run"""

    parser = argparse.ArgumentParser()
    parser.add_argument('tasksdb_filename')
    args = parser.parse_args(arg_list)

    if not os.path.isfile(args.tasksdb_filename):
        raise IOError('Tasks db file not found at: %s' % args.tasksdb_filename)

    tasks, engine_number_map = get_engine_data(args.tasksdb_filename)
    plot_usage(tasks, engine_number_map)
    plot_duration_histogram(tasks)

    plt.show()


def main():
    """Main"""

    run(sys.argv[1:])


if __name__ == '__main__':
    main()



"""Tests for ephys.mechanisms"""

import string
import random
import json
import difflib

import nose.tools as nt
from nose.plugins.attrib import attr

import utils
from bluepyopt import ephys
import bluepyopt.ephys.examples.simplecell as simplecell

from bluepyopt.ephys.serializer import instantiator

simple_cell = simplecell.cell_model
simple_cell.freeze(simplecell.default_param_values)
sim = simplecell.nrn_sim


@attr('unit')
def test_mechanism_serialize():
    """ephys.mechanisms: Testing serialize"""
    mech = utils.make_mech()
    serialized = mech.to_dict()
    nt.assert_true(isinstance(json.dumps(serialized), str))
    deserialized = instantiator(serialized)
    nt.assert_true(isinstance(deserialized, ephys.mechanisms.NrnMODMechanism))


@attr('unit')
def test_nrnmod_instantiate():
    """ephys.mechanisms: Testing insert mechanism"""

    test_mech = ephys.mechanisms.NrnMODMechanism(
        'test.pas',
        suffix='pas',
        locations=[simplecell.somatic_loc])

    nt.assert_equal(str(test_mech), "test.pas: pas at ['somatic']")

    simple_cell.instantiate(sim=sim)

    test_mech.instantiate(sim=sim, icell=simple_cell.icell)
    test_mech.destroy(sim=sim)

    simple_cell.destroy(sim=sim)

    nt.assert_raises(TypeError, ephys.mechanisms.NrnMODMechanism,
                     'test.pas',
                     suffix='pas',
                     prefix='pas',
                     locations=[simplecell.somatic_loc])

    test_mech = ephys.mechanisms.NrnMODMechanism(
        'test.pas',
        prefix='pas',
        locations=[simplecell.somatic_loc])

    nt.assert_equal(test_mech.suffix, 'pas')

    test_mech.prefix = 'pas2'
    nt.assert_equal(test_mech.suffix, 'pas2')

    test_mech = ephys.mechanisms.NrnMODMechanism(
        'unknown',
        suffix='unknown',
        locations=[simplecell.somatic_loc])

    simple_cell.instantiate(sim=sim)

    nt.assert_raises(
        ValueError,
        test_mech.instantiate,
        sim=sim,
        icell=simple_cell.icell)

    test_mech.destroy(sim=sim)
    simple_cell.destroy(sim=sim)


def compare_strings(s1, s2):
    """Compare two strings"""

    diff = list(difflib.unified_diff(s1.splitlines(1), s2.splitlines(1)))

    if len(diff) > 0:
        print(''.join(diff))
        return False
    else:
        return True


@attr('unit')
def test_nrnmod_reinitrng_block():
    """ephys.mechanisms: Testing reinitrng_block"""

    test_mech = ephys.mechanisms.NrnMODMechanism(
        'stoch',
        suffix='Stoch',
        locations=[simplecell.somatic_loc])

    block = test_mech.generate_reinitrng_hoc_block()
    expected_block = '    forsec somatic { deterministic_Stoch = 1 }\n'

    nt.assert_true(compare_strings(block, expected_block))

    test_mech = ephys.mechanisms.NrnMODMechanism(
        'stoch',
        suffix='Stoch',
        deterministic=False,
        locations=[simplecell.somatic_loc])

    block = test_mech.generate_reinitrng_hoc_block()

    expected_block = \
        """    forsec somatic {
        for (x, 0) {
            setdata_Stoch(x)
            sf.tail(secname(), "\\\\.", name)
            sprint(full_str, "%s.%.19g", name, x)
            if (channel_seed_set) {
                setRNG_Stoch(gid, hash_str(full_str), channel_seed)
            } else {
                setRNG_Stoch(gid, hash_str(full_str))
            }
        }
    }
"""

    nt.assert_true(compare_strings(block, expected_block))


@attr('unit')
def test_nrnmod_determinism():
    """ephys.mechanisms: Testing determinism"""

    test_mech = ephys.mechanisms.NrnMODMechanism(
        'pas',
        suffix='pas',
        deterministic=False,
        locations=[simplecell.somatic_loc])

    simple_cell.instantiate(sim=sim)

    nt.assert_raises(
        TypeError,
        test_mech.instantiate,
        sim=sim,
        icell=simple_cell.icell)
    test_mech.destroy(sim=sim)

    simple_cell.destroy(sim=sim)


@attr('unit')
def test_pprocess_instantiate():
    """ephys.mechanisms: Testing insert point process"""

    test_pprocess = ephys.mechanisms.NrnMODPointProcessMechanism(
        name='expsyn',
        suffix='ExpSyn',
        locations=[simplecell.somacenter_loc])

    nt.assert_equal(
        str(test_pprocess),
        "expsyn: ExpSyn at ['somatic[0](0.5)']")

    simple_cell.instantiate(sim=sim)

    nt.assert_equal(test_pprocess.pprocesses, None)

    test_pprocess.instantiate(sim=sim, icell=simple_cell.icell)
    nt.assert_equal(len(test_pprocess.pprocesses), 1)
    pprocess = test_pprocess.pprocesses[0]

    nt.assert_true(hasattr(pprocess, 'tau'))
    test_pprocess.destroy(sim=sim)

    nt.assert_equal(test_pprocess.pprocesses, None)

    simple_cell.destroy(sim=sim)

    test_pprocess = ephys.mechanisms.NrnMODPointProcessMechanism(
        name='expsyn',
        suffix='Exp',
        locations=[simplecell.somacenter_loc])

    simple_cell.instantiate(sim=sim)

    nt.assert_raises(
        AttributeError,
        test_pprocess.instantiate,
        sim=sim,
        icell=simple_cell.icell)

    test_pprocess.destroy(sim=sim)
    simple_cell.destroy(sim=sim)


@attr('unit')
def test_string_hash_functions():
    """ephys.mechanisms: Testing string hash function"""

    n_of_strings = 100
    max_size = 50

    random.seed(1)

    test_strings = ['', 'a']
    test_strings += [''.join
                     (random.choice
                      (string.printable)
                      for _ in range(random.choice(range(max_size))))
                     for _ in range(n_of_strings)]
    hashes_py = [
        ephys.mechanisms.NrnMODMechanism.hash_py
        (test_string) for test_string in test_strings]
    hashes_hoc = [
        ephys.mechanisms.NrnMODMechanism.hash_hoc
        (test_string, simplecell.nrn_sim) for test_string in test_strings]

    nt.assert_equal(hashes_py, hashes_hoc)
    nt.assert_equal(hashes_py[:2], [0.0, 97.0])



"""selIBEA tests"""

import nose.tools as nt

import deap
import numpy

import bluepyopt.deapext
from bluepyopt.deapext.tools.selIBEA \
    import (_calc_fitness_components, _mating_selection,)
from deapext_test_utils import make_mock_population

from nose.plugins.attrib import attr


@attr('unit')
def test_calc_fitness_components():
    """deapext.selIBEA: test calc_fitness_components"""
    KAPPA = 0.05
    population = make_mock_population()

    components = _calc_fitness_components(population, kappa=KAPPA)

    expected = numpy.array(
        [
            [1.00000000e+00, 4.30002298e-05, 4.26748513e-09, 2.06115362e-09,
             9.71587289e-03],
            [5.11484499e-09, 1.00000000e+00, 2.02317572e-07, 4.79335491e-05,
             3.52720088e-08],
            [6.75130710e-07, 1.23735078e+00, 1.00000000e+00, 2.77149617e-01,
             8.37712763e-06],
            [2.06115362e-09, 3.04444453e-04, 8.15288827e-08, 1.00000000e+00,
             2.06115362e-09],
            [2.06115362e-09, 6.75565231e-04, 4.39228177e-07, 2.12142918e-07,
             1.00000000e+00]
        ])

    nt.assert_true(numpy.allclose(expected, components))


@attr('unit')
def test_mating_selection():
    """deapext.selIBEA: test mating selection"""

    PARENT_COUNT = 10
    population = make_mock_population()
    parents = _mating_selection(population, PARENT_COUNT, 5)
    nt.assert_equal(len(parents), PARENT_COUNT)
    expected = [1, 1, 1, 1, 1, 0, 1, 0, 0, 0]
    nt.assert_equal(expected, [ind.ibea_fitness for ind in parents])


@attr('unit')
def test_selibea_init():
    """deapext.selIBEA: test selIBEA init"""

    deap.creator.create('fit', deap.base.Fitness, weights=(-1.0,))
    deap.creator.create(
        'ind',
        numpy.ndarray,
        fitness=deap.creator.__dict__['fit'])

    numpy.random.seed(1)

    population = [deap.creator.__dict__['ind'](x)
                  for x in numpy.random.uniform(0, 1,
                                                (10, 2))]

    for ind in population:
        ind.fitness.values = (numpy.random.uniform(0, 1), )

    mu = 5
    parents = bluepyopt.deapext.tools.selIBEA(population, mu)

    nt.assert_equal(len(parents), mu)
